import chex
import jax
import jax.numpy as jnp
import jax.random as jr
import numpyro
import numpyro.distributions as dist
import pytest
from numpyro.contrib.tfp import distributions as tfd
from numpyro.distributions import constraints

from gpjax.gps import Prior
from gpjax.interfaces.numpyro import (
    add_constraints,
    add_priors,
    numpyro_dict_params,
    numpyro_marginal_ll,
)
from gpjax.kernels import RBF
from gpjax.likelihoods import Gaussian
from gpjax.parameters import initialise
from gpjax.types import Dataset

seed = 123
KEY = jr.PRNGKey(seed)

# TODO: test conjugate posterior
def _get_conjugate_posterior_params() -> dict:
    kernel = RBF()
    prior = Prior(kernel=kernel)
    lik = Gaussian()
    posterior = prior * lik
    params = initialise(posterior)
    return params, posterior


def _gen_training_data(n_samples, n_features, n_latents):

    x = jr.normal(key=KEY, shape=(n_samples, n_features))
    y = jr.normal(key=KEY, shape=(n_samples, n_latents))

    return Dataset(X=x, y=y)


def test_numpyro_marginal_ll_params():

    # create sample data
    ds = _gen_training_data(10, 10, 2)

    # initialize parameters
    params, posterior = _get_conjugate_posterior_params()

    # convert to numpyro-style params
    numpyro_params = numpyro_dict_params(params)

    # initialize numpyro-style GP model
    npy_model = numpyro_marginal_ll(posterior, numpyro_params)

    # do one forward pass with context
    with numpyro.handlers.seed(rng_seed=KEY):
        model_params = numpyro.handlers.trace(npy_model).get_trace(ds)

    assert set(numpyro_params) <= set(model_params)


@pytest.mark.parametrize(
    "n_samples",
    [1, 10, 100],
)
@pytest.mark.parametrize(
    "n_features",
    [1, 10, 100],
)
@pytest.mark.parametrize(
    "n_latents",
    [
        1,
    ],
)
def test_numpyro_marginal_ll_params_shape(n_samples, n_features, n_latents):

    # create sample data
    ds = _gen_training_data(n_samples, n_features, n_latents)

    # initialize parameters
    params, posterior = _get_conjugate_posterior_params()

    # convert to numpyro-style params
    numpyro_params = numpyro_dict_params(params)

    # initialize numpyro-style GP model
    npy_model = numpyro_marginal_ll(posterior, numpyro_params)

    # do one forward pass with context
    with numpyro.handlers.seed(rng_seed=KEY):
        pred = npy_model(ds)

        chex.assert_equal_shape([ds.y.squeeze(), pred])


@pytest.mark.parametrize(
    "n_samples",
    [1, 10, 100],
)
@pytest.mark.parametrize(
    "n_features",
    [1, 10, 100],
)
@pytest.mark.parametrize(
    "n_latents",
    [
        1,
    ],
)
@pytest.mark.parametrize(
    "dtype",
    [
        jnp.float32,
        jnp.float64,
    ],
)
def test_numpyro_marginal_ll_params_type(n_samples, n_features, n_latents, dtype):

    # create sample data
    ds = _gen_training_data(n_samples, n_features, n_latents)

    # convert to dtype
    ds = jax.tree_util.tree_map(lambda x: x.astype(dtype), ds)

    # initialize parameters
    params, posterior = _get_conjugate_posterior_params()

    # convert to numpyro-style params
    numpyro_params = numpyro_dict_params(params)

    # initialize numpyro-style GP model
    npy_model = numpyro_marginal_ll(posterior, numpyro_params)

    # do one forward pass with context
    with numpyro.handlers.seed(rng_seed=KEY):
        pred = npy_model(ds)

        chex.assert_equal(pred.dtype, ds.y.dtype)


def test_numpyro_marginal_ll_numpyro_priors():

    # create sample data
    ds = _gen_training_data(10, 10, 2)

    # initialize parameters
    params, posterior = _get_conjugate_posterior_params()

    # convert to numpyro-style params
    numpyro_params = numpyro_dict_params(params)

    # convert to priors
    numpyro_params = add_priors(numpyro_params, dist.LogNormal(0.0, 10.0))

    # initialize numpyro-style GP model
    npy_model = numpyro_marginal_ll(posterior, numpyro_params)

    # do one forward pass with context
    with numpyro.handlers.seed(rng_seed=KEY):
        model_params = numpyro.handlers.trace(npy_model).get_trace(ds)

    assert set(numpyro_params) <= set(model_params)


@pytest.mark.parametrize(
    "n_samples",
    [1, 10, 100],
)
@pytest.mark.parametrize(
    "n_features",
    [1, 10, 100],
)
@pytest.mark.parametrize(
    "n_latents",
    [
        1,
    ],
)
def test_numpyro_marginal_ll_numpyro_priors_shape(n_samples, n_features, n_latents):

    # create sample data
    ds = _gen_training_data(n_samples, n_features, n_latents)

    # initialize parameters
    params, posterior = _get_conjugate_posterior_params()

    # convert to numpyro-style params
    numpyro_params = numpyro_dict_params(params)

    # convert to priors
    numpyro_params = add_priors(numpyro_params, dist.LogNormal(0.0, 10.0))

    # initialize numpyro-style GP model
    npy_model = numpyro_marginal_ll(posterior, numpyro_params)

    # do one forward pass with context
    with numpyro.handlers.seed(rng_seed=KEY):
        pred = npy_model(ds)

        chex.assert_equal_shape([pred, ds.y.squeeze()])


@pytest.mark.parametrize(
    "n_samples",
    [1, 10, 100],
)
@pytest.mark.parametrize(
    "n_features",
    [1, 10, 100],
)
@pytest.mark.parametrize(
    "n_latents",
    [
        1,
    ],
)
@pytest.mark.parametrize(
    "dtype",
    [
        jnp.float32,
        jnp.float64,
    ],
)
def test_numpyro_marginal_ll_numpyro_priors_type(n_samples, n_features, n_latents, dtype):

    # create sample data
    ds = _gen_training_data(n_samples, n_features, n_latents)

    # convert to tyle
    ds = jax.tree_util.tree_map(lambda x: x.astype(dtype), ds)

    # initialize parameters
    params, posterior = _get_conjugate_posterior_params()

    # convert to numpyro-style params
    numpyro_params = numpyro_dict_params(params)

    # convert to priors
    numpyro_params = add_priors(numpyro_params, dist.LogNormal(0.0, 10.0))

    # initialize numpyro-style GP model
    npy_model = numpyro_marginal_ll(posterior, numpyro_params)

    # do one forward pass with context
    with numpyro.handlers.seed(rng_seed=KEY):
        pred = npy_model(ds)

        chex.assert_equal(pred.dtype, ds.y.dtype)


@pytest.mark.parametrize(
    "n_samples",
    [1, 10, 100],
)
@pytest.mark.parametrize(
    "n_features",
    [1, 10, 100],
)
@pytest.mark.parametrize(
    "n_latents",
    [
        1,
    ],
)
@pytest.mark.parametrize(
    "dtype",
    [
        jnp.float32,
        jnp.float64,
    ],
)
def test_numpyro_marginal_ll_tfp_priors_type(n_samples, n_features, n_latents, dtype):

    # create sample data
    ds = _gen_training_data(n_samples, n_features, n_latents)

    # convert to tyle
    ds = jax.tree_util.tree_map(lambda x: x.astype(dtype), ds)

    # initialize parameters
    params, posterior = _get_conjugate_posterior_params()

    # convert to numpyro-style params
    numpyro_params = numpyro_dict_params(params)

    # convert to priors
    numpyro_params = add_priors(numpyro_params, tfd.LogNormal(0.0, 10.0))

    # initialize numpyro-style GP model
    npy_model = numpyro_marginal_ll(posterior, numpyro_params)

    # do one forward pass with context
    with numpyro.handlers.seed(rng_seed=KEY):
        pred = npy_model(ds)

        chex.assert_equal(pred.dtype, ds.y.dtype)



from .config import get_defaults
from .gps import Prior
from .kernels import RBF
from .likelihoods import Gaussian
from .objectives import marginal_ll
from .parameters import build_all_transforms, initialise
from .predict import mean, variance
from .sampling import random_variable, sample
from .types import Dataset
from .utils import as_constant
from .mean_functions import Zero


from typing import Optional, Union

import jax.numpy as jnp
from chex import dataclass
from jax.interpreters.ad import JVPTracer
from jax.interpreters.partial_eval import DynamicJaxprTracer
from jax.interpreters.pxla import ShardedDeviceArray

NoneType = type(None)
Array = (
    jnp.ndarray,
    ShardedDeviceArray,
    jnp.DeviceArray,
    JVPTracer,
    DynamicJaxprTracer,
)
Arrays = Union[jnp.ndarray, ShardedDeviceArray, jnp.DeviceArray, JVPTracer, DynamicJaxprTracer]


@dataclass(repr=False)
class Dataset:
    X: Arrays
    y: Arrays = None

    def __repr__(self) -> str:
        return f"- Number of datapoints: {self.X.shape[0]}\n- Dimension: {self.X.shape[1]}"

    @property
    def n(self) -> int:
        return self.X.shape[0]

    @property
    def in_dim(self) -> int:
        return self.X.shape[1]

    @property
    def out_dim(self) -> int:
        return self.y.shape[1]


def verify_dataset(ds: Dataset) -> None:
    assert (
        ds.X.ndim == 2
    ), f"2-dimensional training inputs are required. Current dimension: {ds.X.ndim}."
    if ds.y is not None:
        assert (
            ds.y.ndim == 2
        ), f"2-dimensional training outputs are required. Current dimension: {ds.y.ndim}."
        assert (
            ds.X.shape[0] == ds.y.shape[0]
        ), f"Number of inputs must equal the number of outputs. \nCurrent counts:\n- X: {ds.X.shape[0]}\n- y: {ds.y.shape[0]}"



# from typing import Iterable, Dict
from collections import Iterable
from copy import deepcopy
from typing import Callable, List, Union

import jax.numpy as jnp
import numpyro
import numpyro.distributions as dist
from chex import Array
from jax.scipy.linalg import cholesky
from multipledispatch import dispatch
from numpyro import module

from gpjax.gps import ConjugatePosterior, NonConjugatePosterior
from gpjax.kernels import gram
from gpjax.likelihoods import link_function
from gpjax.types import Dataset
from gpjax.utils import I

numpyro_constraint = numpyro.distributions.constraints.Constraint
numpyro_priors = numpyro.distributions.Distribution


def numpyro_dict_params(params: dict) -> List:
    """Function to go from standard dictionary to params"""
    numpyro_params = {}
    for ikey, iparam in params.items():

        if isinstance(iparam, jnp.DeviceArray) or isinstance(iparam, float):
            numpyro_params[ikey] = {
                "param_type": "param",
                "init_value": iparam,
                "constraint": dist.constraints.positive,
            }

        elif isinstance(iparam, numpyro_priors):
            numpyro_params[ikey] = {
                "param_type": "prior",
                "prior": iparam,
            }

        else:
            raise ValueError(f"Unrecognized param input: {iparam}")

    return numpyro_params


@dispatch(dict, numpyro_constraint)
def add_constraints(params: dict, constraint: numpyro_constraint):
    new_params = deepcopy(params)
    for ivariable in params.keys():
        if new_params[ivariable]["param_type"] == "param":
            new_params[ivariable]["constraint"] = constraint
    return new_params


@dispatch(dict, dict)
def add_constraints(params: dict, constraint: Iterable):
    new_params = deepcopy(params)

    for ivariable, iconstraint in constraint.items():
        if new_params[ivariable]["param_type"] == "param":
            new_params[ivariable]["constraint"] = iconstraint
    return new_params


@dispatch(dict, str, numpyro_constraint)
def add_constraints(params: dict, variable: str, constraint: numpyro_constraint):
    new_params = deepcopy(params)
    if new_params[variable]["param_type"] == "param":
        new_params[variable]["constraint"] = constraint
    return new_params


@dispatch(dict, dict)
def add_priors(params: dict, priors: dict):
    new_params = deepcopy(params)
    for ivariable, iprior in priors.items():
        new_params[ivariable] = {"param_type": "prior", "prior": iprior}
    return new_params


@dispatch(dict, str, numpyro_priors)
def add_priors(params: dict, variable: str, priors: numpyro_priors):
    new_params = deepcopy(params)
    new_params[variable] = {"param_type": "prior", "prior": priors}
    return new_params


@dispatch(dict, numpyro_priors)
def add_priors(params: dict, priors: numpyro_priors):
    new_params = deepcopy(params)
    for ivariable in new_params.keys():
        new_params[ivariable] = {"param_type": "prior", "prior": priors}
    return new_params


@dispatch(ConjugatePosterior, dict)
def numpyro_marginal_ll(gp: ConjugatePosterior, numpyro_params: dict) -> Callable:
    def mll(ds: Dataset):

        x, y = ds.X, ds.y

        params = {}

        for iname, iparam in numpyro_params.items():
            if iparam["param_type"] == "prior":
                params[iname] = numpyro.sample(name=iname, fn=iparam["prior"])
            else:
                params[iname] = numpyro.param(
                    name=iname,
                    init_value=iparam["init_value"],
                    constraint=iparam["constraint"],
                )
        # get mean function
        mu = gp.prior.mean_function(x)

        # covariance function
        gram_matrix = gram(gp.prior.kernel, x, params)
        gram_matrix += params["obs_noise"] * I(x.shape[0])

        # scale triangular matrix
        L = cholesky(gram_matrix, lower=True)
        return numpyro.sample(
            "y",
            dist.MultivariateNormal(loc=mu, scale_tril=L),
            obs=y.squeeze(),
        )

    return mll



from typing import Tuple

import tensorflow_probability.substrates.jax.bijectors as tfb
from ml_collections import ConfigDict


def get_defaults():
    config = ConfigDict()
    # Covariance matrix stabilising jitter
    config.jitter = 1e-6

    # Default bijections
    config.transformations = transformations = ConfigDict()
    transformations.positive_transform = tfb.Softplus()
    transformations.identity_transform = tfb.Identity()

    # Default parameter transforms
    transformations.lengthscale = "positive_transform"
    transformations.variance = "positive_transform"
    transformations.obs_noise = "positive_transform"
    transformations.latent = "identity_transform"
    transformations.basis_fns = "identity_transform"
    return config


def add_parameter(config: ConfigDict, bijection_tuple: Tuple[str, tfb.Bijector]) -> ConfigDict:
    param_name, bijection = bijection_tuple
    lookup_name = f"custom_{param_name}"
    config.transformations[lookup_name] = bijection
    config.transformations[param_name] = lookup_name
    return config



import chex
import jax.numpy as jnp
import jax.random as jr
import numpy as np
import numpyro
import numpyro.distributions as dist
import pytest
from numpyro.distributions import constraints

from gpjax.gps import Prior
from gpjax.interfaces.numpyro import add_constraints, numpyro_dict_params
from gpjax.kernels import RBF
from gpjax.likelihoods import Gaussian
from gpjax.parameters import initialise


# TODO: test conjugate posterior
def _get_conjugate_posterior_params() -> dict:
    kernel = RBF()
    prior = Prior(kernel=kernel)
    lik = Gaussian()
    posterior = prior * lik
    params = initialise(posterior)
    return params


@pytest.mark.parametrize(
    "input_types",
    [
        str(1.0),
        np.array(1.0),
        None,
    ],
)
def test_numpyro_dict_params_defaults_nullcase(input_types):

    demo_params = {
        "lengthscale": input_types,
        "variance": input_types,
        "obs_noise": input_types,
    }
    with pytest.raises(ValueError):

        numpyro_dict_params(demo_params)


def test_numpyro_dict_params_defaults_array():

    demo_params = {
        "lengthscale": jnp.array(1.0),
        "variance": jnp.array(1.0),
        "obs_noise": jnp.array(1.0),
    }

    numpyro_params = numpyro_dict_params(demo_params)

    assert set(numpyro_params) == set(demo_params.keys())
    for ikey, iparam in demo_params.items():
        # check keys exist for param
        assert set(numpyro_params[ikey].keys()) == set(("init_value", "constraint", "param_type"))
        # check init value is the same as initial value
        chex.assert_equal(numpyro_params[ikey]["init_value"], iparam)
        # check default constraint is positive
        chex.assert_equal(numpyro_params[ikey]["constraint"], constraints.positive)
        # check if param type is param
        chex.assert_equal(numpyro_params[ikey]["param_type"], "param")

    # check we didn't modify original dictionary
    chex.assert_equal(
        demo_params,
        {
            "lengthscale": jnp.array(1.0),
            "variance": jnp.array(1.0),
            "obs_noise": jnp.array(1.0),
        },
    )


def test_numpyro_dict_params_defaults_float():

    demo_params = {
        "lengthscale": 1.0,
        "variance": 1.0,
        "obs_noise": 1.0,
    }

    numpyro_params = numpyro_dict_params(demo_params)

    assert set(numpyro_params) == set(demo_params.keys())
    for ikey, iparam in demo_params.items():
        # check keys exist for param
        assert set(numpyro_params[ikey].keys()) == set(("init_value", "constraint", "param_type"))
        # check init value is the same as initial value
        chex.assert_equal(numpyro_params[ikey]["init_value"], iparam)
        # check default constraint is positive
        chex.assert_equal(numpyro_params[ikey]["constraint"], constraints.positive)
        # check if param type is param
        chex.assert_equal(numpyro_params[ikey]["param_type"], "param")

    # check we didn't modify original dictionary
    chex.assert_equal(
        demo_params,
        {
            "lengthscale": 1.0,
            "variance": 1.0,
            "obs_noise": 1.0,
        },
    )


@pytest.mark.parametrize(
    "constraint",
    [
        constraints.positive,
        constraints.real,
        constraints.softplus_positive,
        constraints.softplus_lower_cholesky,
    ],
)
def test_numpyro_add_constraints_all(constraint):

    gpjax_params = _get_conjugate_posterior_params()
    numpyro_params = numpyro_dict_params(gpjax_params)

    # add constraint
    new_numpyro_params = add_constraints(numpyro_params, constraint)
    for iparams in new_numpyro_params.values():

        # check if constraint in new dictionary
        chex.assert_equal(iparams["constraint"], constraint)

    # check we didn't modify original dictionary
    chex.assert_equal(gpjax_params, _get_conjugate_posterior_params())


@pytest.mark.parametrize(
    "variable",
    ["lengthscale", "obs_noise", "variance"],
)
@pytest.mark.parametrize(
    "constraint",
    [
        constraints.positive,
        constraints.real,
        constraints.softplus_positive,
        constraints.softplus_lower_cholesky,
    ],
)
def test_numpyro_add_constraints_str(variable, constraint):

    gpjax_params = _get_conjugate_posterior_params()
    numpyro_params = numpyro_dict_params(gpjax_params)

    # add constraint
    new_numpyro_params = add_constraints(numpyro_params, variable, constraint)

    # check if constraint in new dictionary
    chex.assert_equal(new_numpyro_params[variable]["constraint"], constraint)

    # check we didn't modify original dictionary
    chex.assert_equal(gpjax_params, _get_conjugate_posterior_params())


@pytest.mark.parametrize(
    "variable",
    ["lengthscale", "obs_noise", "variance"],
)
@pytest.mark.parametrize(
    "constraint",
    [
        constraints.positive,
        constraints.real,
        constraints.softplus_positive,
        constraints.softplus_lower_cholesky,
    ],
)
def test_numpyro_add_constraints_dict(variable, constraint):

    gpjax_params = _get_conjugate_posterior_params()
    numpyro_params = numpyro_dict_params(gpjax_params)

    # create new dictionary
    new_param_dict = {str(variable): constraint}

    # add constraint
    new_numpyro_params = add_constraints(numpyro_params, new_param_dict)

    # check if constraint in new dictionary
    chex.assert_equal(new_numpyro_params[variable]["constraint"], constraint)

    # check we didn't modify original dictionary
    chex.assert_equal(gpjax_params, _get_conjugate_posterior_params())
